@@ -37,17 +37,22 @@
 #include <getopt.h>
 #include <string.h>
 #include <errno.h>
+#include <io.h>
 #include "imagehelper.h"
 #include "rebase-db.h"
+#include <versionhelpers.h> /* requires <windows.h> */
 
 BOOL save_image_info ();
 BOOL load_image_info ();
 BOOL merge_image_info ();
 BOOL collect_image_info (const char *pathname);
 void print_image_info ();
+static void print_overlapped ();
 BOOL rebase (const char *pathname, ULONG64 *new_image_base, BOOL down_flag);
 void parse_args (int argc, char *argv[]);
 unsigned long long string_to_ulonglong (const char *string);
+static int compactos_get_algorithm (const char *pathname);
+static int compactos_compress_file (const char *pathname, int algorithm);
 void usage ();
 void help ();
 BOOL is_rebaseable (const char *pathname);
@@ -259,9 +264,15 @@ main (int argc, char *argv[])
       ULONG64 new_image_base = image_base;
       for (i = 0; i < img_info_size; ++i)
 	{
+	  int compactos_algorithm
+	      = compactos_get_algorithm (img_info_list[i].name);
 	  status = rebase (img_info_list[i].name, &new_image_base, down_flag);
 	  if (!status)
 	    return 2;
+	  /* Reapply previous compression. */
+	  if (compactos_algorithm >= 0)
+	    compactos_compress_file (img_info_list[i].name,
+				     compactos_algorithm);
 	}
     }
   else
@@ -269,6 +280,9 @@ main (int argc, char *argv[])
       /* Rebase with database support. */
       BOOL header;
 
+#if defined(__CYGWIN__)
+      int compactos_algorithm = compactos_get_algorithm ("/bin/cygwin1.dll");
+#endif
       if (merge_image_info () < 0)
 	return 2;
       status = TRUE;
@@ -279,6 +293,14 @@ main (int argc, char *argv[])
 	    status = rebase (img_info_list[i].name, &new_image_base, FALSE);
 	    if (status)
 	      img_info_list[i].flag.needs_rebasing = 0;
+#if defined(__CYGWIN__)
+	    /* If Cygwin DLL is compressed, assume setup was used with option
+	       --compact-os.  Align compression with Cygwin DLL. */
+	    if (compactos_algorithm >= 0
+		&& compactos_compress_file (img_info_list[i].name,
+					    compactos_algorithm) < 0)
+	      compactos_algorithm = -1;
+#endif
 	  }
       for (header = FALSE, i = 0; i < img_info_size; ++i)
 	if (img_info_list[i].flag.cannot_rebase == 1)
@@ -302,6 +324,10 @@ main (int argc, char *argv[])
 	      }
 	    fprintf (stderr, "  %s\n", img_info_list[i].name);
 	  }
+      /* Print list of DLLs which still overlap.  This could occur if DLLs are
+	 not rebaseable or if --merge-files is used incorrectly. */
+      if (img_info_size)
+	print_overlapped ();
       if (save_image_info () < 0)
 	return 2;
     }
@@ -589,6 +615,7 @@ set_cannot_rebase (img_info_t *img)
    * is set to 2 on loading the database entries */
   if (img->flag.cannot_rebase <= 1 )
     {
+      /* This also removes Compact OS compression. */
       int fd = open (img->name, O_WRONLY);
       if (fd < 0)
 	img->flag.cannot_rebase = 1;
@@ -603,6 +630,7 @@ merge_image_info ()
 {
   int i, end;
   img_info_t *match;
+  BOOL sorted;
   ULONG64 floating_image_base;
 
   /* Sort new files from command line by name. */
@@ -682,6 +710,17 @@ merge_image_info ()
 	      if (verbose)
 		fprintf (stderr, "rebasing %s because not in database yet\n", img_info_list[i].name);
 	    }
+	  else if (img_info_list[i].flag.needs_rebasing)
+	    {
+	      /* Not in database yet and not rebaseable.  Add without rebasing or
+	         skip if --oblivious is active. */
+	      img_info_list[i].flag.needs_rebasing = 0;
+	      if (verbose)
+		fprintf (stderr, "not rebasing %s because file is not writable\n",
+			 img_info_list[i].name);
+	      /* FIXME: Overlaps of DLLs in the database with this DLL will
+	         not be detected below. */
+	    }
 	}
     }
   if (!img_info_rebase_start || force_rebase_flag)
@@ -692,7 +731,12 @@ merge_image_info ()
 	{
 	  /* Test DLLs already in database for writability. */
 	  if (i < img_info_rebase_start)
-	    set_cannot_rebase (&img_info_list[i]);
+	    {
+	      set_cannot_rebase (&img_info_list[i]);
+	      if (img_info_list[i].flag.cannot_rebase == 1 && verbose)
+		fprintf (stderr, "not rebasing %s because file is not writable\n",
+			 img_info_list[i].name);
+	    }
 	  if (!img_info_list[i].flag.cannot_rebase)
 	    {
 	      img_info_list[i].base = 0;
@@ -703,47 +747,71 @@ merge_image_info ()
       img_info_rebase_start = 0;
     }
 
-  /* Now sort the old part of the list by base address. */
-  if (img_info_rebase_start)
-    qsort (img_info_list, img_info_rebase_start, sizeof (img_info_t),
-	   img_info_cmp);
-  /* Perform several tests on the information fetched from the database
-     to match with reality. */
-  for (i = 0; i < img_info_rebase_start; ++i)
+  for (sorted = FALSE; img_info_rebase_start && !sorted; )
     {
-      ULONG64 cur_base;
-      ULONG cur_size, slot_size;
+      char overlaps[img_info_rebase_start];
+      memset(overlaps, 0, img_info_rebase_start);
+      /* Now sort the old part of the list by base address. */
+      qsort (img_info_list, img_info_rebase_start, sizeof (img_info_t),
+	     img_info_cmp);
+      /* Perform several tests on the information fetched from the database
+	 to match with reality. */
+      for (sorted = TRUE, i = 0; sorted && i < img_info_rebase_start; ++i)
+	{
+	  ULONG64 cur_base, cur_base_orig;
+	  ULONG cur_size, slot_size;
 
-      /* Files with the needs_rebasing or cannot_rebase flags set have been
-	 checked already. */
-      if (img_info_list[i].flag.needs_rebasing
-      	  || img_info_list[i].flag.cannot_rebase)
-	continue;
-      /* Check if the files in the old list still exist.  Drop non-existant
-	 or unaccessible files. */
-      if (access (img_info_list[i].name, F_OK) == -1
-	  || !GetImageInfos64 (img_info_list[i].name, NULL,
-			       &cur_base, &cur_size))
-	{
-	  free (img_info_list[i].name);
-	  memmove (img_info_list + i, img_info_list + i + 1,
-		   (img_info_size - i - 1) * sizeof (img_info_t));
-	  --img_info_rebase_start;
-	  --img_info_size;
-	  continue;
-	}
-      slot_size = roundup2 (cur_size, ALLOCATION_SLOT);
-      if (set_cannot_rebase (&img_info_list[i]))
-	img_info_list[i].base = cur_base;
-      else
-	{
+	  /* If file is not rebaseable and neither --oblivious nor --merge-files
+	     is active, search forward for possible overlaps. */
+	  if (img_info_list[i].flag.cannot_rebase == 1)
+	    {
+	      int j;
+	      /* If cannot_rebase is set, base and size are already valid. */
+	      cur_base = img_info_list[i].base;
+	      slot_size = img_info_list[i].slot_size;
+	      for (j = i + 1; j < img_info_rebase_start; ++j)
+		{
+		  if (img_info_list[j].flag.cannot_rebase)
+		    continue; /* Also not rebaseable. */
+		  if (img_info_list[j].base == 0)
+		    continue; /* Will already be rebased from scratch. */
+		  if (cur_base + slot_size + offset <= img_info_list[j].base)
+		    break; /* This and any further DLLs do not overlap. */
+		  /* Select rebase from scratch when outer loop arrives at j. */
+		  overlaps[j] = 1;
+		}
+	    }
+
+	  /* Files with the needs_rebasing or cannot_rebase flags set have been
+	     checked already. */
+	  if (img_info_list[i].flag.needs_rebasing
+	      || img_info_list[i].flag.cannot_rebase)
+	    continue;
+	  /* Check if the files in the old list still exist.  Drop non-existant
+	     or unaccessible files. */
+	  if (access (img_info_list[i].name, F_OK) == -1
+	      || !GetImageInfos64 (img_info_list[i].name, NULL,
+				   &cur_base, &cur_size))
+	    {
+	      free (img_info_list[i].name);
+	      memmove (img_info_list + i, img_info_list + i + 1,
+		       (img_info_size - i - 1) * sizeof (img_info_t));
+	      memmove (overlaps + i, overlaps + i + 1, img_info_size - i - 1);
+	      --img_info_rebase_start;
+	      --img_info_size;
+	      --i;
+	      continue;
+	    }
+	  slot_size = roundup2 (cur_size, ALLOCATION_SLOT);
+	  cur_base_orig = cur_base;
 	  /* If the file has been reinstalled, try to rebase to the same address
 	     in the first place. */
 	  if (cur_base != img_info_list[i].base)
 	    {
 	      img_info_list[i].flag.needs_rebasing = 1;
 	      if (verbose)
-		fprintf (stderr, "rebasing %s because it's base has changed (due to being reinstalled?)\n", img_info_list[i].name);
+		fprintf (stderr, "rebasing %s because it's base has changed "
+			 "(due to being reinstalled?)\n", img_info_list[i].name);
 	      /* Set cur_base to the old base to simplify subsequent tests. */
 	      cur_base = img_info_list[i].base;
 	    }
@@ -754,16 +822,17 @@ merge_image_info ()
 	    {
 	      img_info_list[i].base = 0;
 	      if (verbose)
-		fprintf (stderr, "rebasing %s because it won't fit in it's old slot without overlapping next DLL\n", img_info_list[i].name);
+		fprintf (stderr, "rebasing %s because it won't fit in it's old slot "
+			 "without overlapping next DLL\n", img_info_list[i].name);
 	    }
-	  /* Does the previous DLL reach into the address space of this
+	  /* Does a previous DLL reach into the address space of this
 	     DLL?  This happens if the previous DLL is not rebaseable. */
-	  else if (i > 0 && cur_base < img_info_list[i - 1].base
-				       + img_info_list[i - 1].slot_size)
+	  else if (overlaps[i])
 	    {
 	      img_info_list[i].base = 0;
 	      if (verbose)
-		fprintf (stderr, "rebasing %s because previous DLL now overlaps\n", img_info_list[i].name);
+		fprintf (stderr, "rebasing %s because a previous non-writable DLL "
+			 "overlaps\n", img_info_list[i].name);
 	    }
 	  /* Does the file match the base address requirements?  If not,
 	     rebase from scratch. */
@@ -772,16 +841,36 @@ merge_image_info ()
 	    {
 	      img_info_list[i].base = 0;
 	      if (verbose)
-		fprintf (stderr, "rebasing %s because it's base address is outside the expected area\n", img_info_list[i].name);
+		fprintf (stderr, "rebasing %s because it's base address is outside the "
+			 "expected area\n", img_info_list[i].name);
 	    }
+	  /* Make sure all DLLs with base address 0 have the needs_rebasing
+	     flag set. */
+	  if (img_info_list[i].base == 0)
+	    img_info_list[i].flag.needs_rebasing = 1;
+	  /* Only check for writability if file needs rebasing to keep
+	     Compact OS compression of unchanged files.  Revert rebase
+	     decision if not writeable. */
+	  if (img_info_list[i].flag.needs_rebasing
+	      && set_cannot_rebase (&img_info_list[i]))
+	    {
+	      img_info_list[i].flag.needs_rebasing = 0;
+	      img_info_list[i].base = cur_base_orig;
+	      if (verbose)
+		fprintf (stderr, "rebasing %s deferred because file is not writable\n",
+			 img_info_list[i].name);
+	      /* List is possibly no longer sorted which could result in false
+	         negative rebase decisions.  Exit loop and restart it with
+		 newly sorted list.  This will also set the overlaps[] array
+		 for this DLL when the restarted loop visits the (possibly new)
+		 index of this DLL.  Infinite retries could not occur because
+		 each turn sets another cannot_rebase flag. */
+	      sorted = FALSE;
+	    }
+	  /* Unconditionally overwrite old with new size. */
+	  img_info_list[i].size = cur_size;
+	  img_info_list[i].slot_size = slot_size;
 	}
-      /* Unconditionally overwrite old with new size. */
-      img_info_list[i].size = cur_size;
-      img_info_list[i].slot_size = slot_size;
-      /* Make sure all DLLs with base address 0 have the needs_rebasing
-	 flag set. */
-      if (img_info_list[i].base == 0)
-	img_info_list[i].flag.needs_rebasing = 1;
     }
   /* The remainder of the function expects img_info_size to be > 0. */
   if (img_info_size == 0)
@@ -1066,6 +1155,39 @@ print_image_info ()
     }
 }
 
+static void
+print_overlapped ()
+{
+  BOOL header;
+  int i;
+  char overlaps[img_info_size];
+  memset (overlaps, 0, img_info_size);
+  qsort (img_info_list, img_info_size, sizeof (img_info_t), img_info_cmp);
+  for (header = FALSE, i = 0; i < img_info_size; ++i)
+    {
+      int j;
+      if (img_info_list[i].flag.needs_rebasing)
+	continue; /* Rebase failed. */
+      for (j = i + 1; j < img_info_size; ++j)
+	{
+	  if (img_info_list[j].flag.needs_rebasing)
+	    continue; /* Rebase failed. */
+	  if (img_info_list[i].base + img_info_list[i].slot_size + offset
+	      <= img_info_list[j].base)
+	    break;
+	  overlaps[i] = overlaps[j] = 1;
+	}
+      if (!overlaps[i])
+	continue;
+      if (!header)
+	{
+	  fputs ("\nThe following DLLs still overlap:\n", stderr);
+	  header = TRUE;
+	}
+      fprintf (stderr, "  %s\n", img_info_list[i].name);
+    }
+}
+
 BOOL
 rebase (const char *pathname, ULONG64 *new_image_base, BOOL down_flag)
 {
@@ -1380,6 +1502,79 @@ string_to_ulonglong (const char *string)
   return number;
 }
 
+static int
+compactos_get_algorithm (const char *pathname)
+{
+  /* Requires Win10. */
+  if (!IsWindows10OrGreater ())
+    return -1;
+
+  int fd = open (pathname, O_RDONLY);
+  if (fd == -1)
+    return -1;
+
+  struct {
+    WOF_EXTERNAL_INFO Wof;
+    FILE_PROVIDER_EXTERNAL_INFO_V1 FileProvider;
+  } wfp = { {0, 0}, {0, 0, 0} };
+
+  int rc;
+  if (!DeviceIoControl ((HANDLE) _get_osfhandle (fd),
+			FSCTL_GET_EXTERNAL_BACKING, NULL, 0,
+			&wfp, sizeof(wfp), NULL, NULL))
+    rc = -1;
+  else
+    rc = (wfp.Wof.Provider == WOF_PROVIDER_FILE ?
+	  wfp.FileProvider.Algorithm : -1);
+
+  close (fd);
+  return rc;
+}
+
+static int
+compactos_compress_file (const char *pathname, int algorithm)
+{
+  /* Do not apply unknown algorithms. */
+  if (!(FILE_PROVIDER_COMPRESSION_XPRESS4K <= algorithm &&
+        algorithm <= FILE_PROVIDER_COMPRESSION_XPRESS16K))
+    return 0;
+
+  int fd = open (pathname, O_RDONLY);
+  if (fd == -1)
+    return -1;
+  HANDLE h = (HANDLE) _get_osfhandle (fd);
+
+  /* Older versions of Win10 set mtime to current time. */
+  FILETIME ft;
+  BOOL ft_valid = GetFileTime (h, NULL, NULL, &ft);
+
+  struct {
+    WOF_EXTERNAL_INFO Wof;
+    FILE_PROVIDER_EXTERNAL_INFO_V1 FileProvider;
+  } wfp;
+  wfp.Wof.Version = WOF_CURRENT_VERSION;
+  wfp.Wof.Provider = WOF_PROVIDER_FILE;
+  wfp.FileProvider.Version = FILE_PROVIDER_CURRENT_VERSION;
+  wfp.FileProvider.Algorithm = algorithm;
+  wfp.FileProvider.Flags = 0;
+
+  int rc;
+  if (!DeviceIoControl (h, FSCTL_SET_EXTERNAL_BACKING, &wfp, sizeof(wfp),
+			NULL, 0, NULL, NULL))
+    rc = (GetLastError() == ERROR_COMPRESSION_NOT_BENEFICIAL ? 0 : -1);
+  else
+    rc = 1;
+
+  if (ft_valid)
+    SetFileTime (h, NULL, NULL, &ft);
+
+  close (fd);
+  if (verbose)
+    printf ("%s: Compact OS algorithm %d %s\n", pathname, algorithm,
+	    (rc < 0 ? "FAILED" : rc == 0 ? "not applied " : "applied"));
+  return rc;
+}
+
 void
 usage ()
 {
